"""
The code that triggers the creation of the necessary code constructs
to realize a sum factorization kernel
"""

from dune.perftool.generation import (barrier,
                                      dump_accumulate_timer,
                                      generator_factory,
                                      get_global_context_value,
                                      globalarg,
                                      instruction,
                                      post_include,
                                      silenced_warning,
                                      transform,
                                      )
from dune.perftool.loopy.buffer import (get_buffer_temporary,
                                        switch_base_storage,
                                        )
from dune.perftool.pdelab.argument import pymbolic_coefficient
from dune.perftool.pdelab.geometry import world_dimension
from dune.perftool.pdelab.spaces import name_lfs, name_lfs_bound
from dune.perftool.options import get_option
from dune.perftool.pdelab.signatures import assembler_routine_name
from dune.perftool.sumfact.permutation import (_sf_permutation_strategy,
                                               _permute_backward,
                                               _permute_forward,
                                               )
from dune.perftool.sumfact.vectorization import attach_vectorization_info
from dune.perftool.sumfact.sumfact import sumfact_iname

import loopy as lp
import numpy as np
import pymbolic.primitives as prim


@generator_factory(item_tags=("sumfactkernel",),
                   context_tags=("kernel",),
                   cache_key_generator=lambda s, **kw: s.cache_key)
def realize_sum_factorization_kernel(sf, insn_dep=frozenset(), outshape=None, direct_input=None, direct_output=None):
    # Unify the insn_dep parameter to be a frozenset
    if isinstance(insn_dep, str):
        insn_dep = frozenset({insn_dep})
    assert isinstance(insn_dep, frozenset)

    # Get the vectorization info. During dry run, this is a now op
#    sf = attach_vectorization_info(sf)
    if get_global_context_value("dry_run", False):
        # During the dry run, we just return the kernel as passed into this
        # function. After the dry run, it can be used to attach information
        # about vectorization.
        return sf, insn_dep

    # Get the instruction dependencies of the sumfact kernel. This variable will be
    # updated throughout this function.
    insn_dep = insn_dep.union(sf.insn_dep)

    # Define some helper constructs that make it easier to write generic code later
    vecindex = () if sf.index is None else (sf.index,)

    # Set up the input for stage 1
    if sf.stage == 1 and not get_option("fastdg"):
        assert sf.coeff_func

        # Get the input temporary!
        input_setup = get_buffer_temporary(sf.buffer,
                                           shape=sf.flat_input_shape,
                                           )

        # Write initial coefficients into buffer
        lfs = name_lfs(sf.element, sf.restriction, sf.component)
        basisiname = sumfact_iname(name_lfs_bound(lfs), "basis")
        container = sf.coeff_func(sf.restriction)
        coeff = pymbolic_coefficient(container, lfs, basisiname)
        assignee = prim.Subscript(prim.Variable(input_setup), (prim.Variable(basisiname),) + vecindex)
        insn_dep = instruction(assignee=assignee,
                               expression=coeff,
                               )

    # Prepare some dim_tags/shapes for later use
    ftags = ",".join(["f"] * sf.length)
    novec_ftags = ftags
    ctags = ",".join(["c"] * sf.length)
    vec_shape = ()
    if sf.vectorized:
        ftags = ftags + ",vec"
        ctags = ctags + ",vec"
        vec_shape = (4,)

    # Measure times and count operations in c++ code
    if get_option("instrumentation_level") >= 4:
        timer_name = assembler_routine_name() + '_kernel' + '_stage{}'.format(sf.stage)
        post_include('HP_DECLARE_TIMER({});'.format(timer_name), filetag='operatorfile')
        dump_accumulate_timer(timer_name)
        insn_dep = frozenset({instruction(code="HP_TIMER_START({});".format(timer_name),
                                          depends_on=insn_dep,
                                          within_inames=frozenset(sf.within_inames))})

    # Put a barrier before the sumfactorization kernel
    insn_dep = frozenset({barrier(depends_on=insn_dep,
                                  within_inames=frozenset(sf.within_inames),
                                  )})

    # Decide in which order we want to process directions in the
    # sumfactorization. A clever ordering can lead to a reduced
    # complexity. This will e.g. happen at faces where we only have
    # one quadratue point m_l=1 if l is the normal direction of the
    # face.
    #
    # Rule of thumb: small m's early and large n's late.
    perm = _sf_permutation_strategy(sf.a_matrices, sf.stage)

    # Permute a_matrices
    a_matrices = _permute_forward(sf.a_matrices, perm)

    # Product of all matrices
    for l, a_matrix in enumerate(a_matrices):
        # Compute the correct shapes of in- and output matrices of this matrix-matrix multiplication
        # and get inames that realize the product.
        inp_shape = (a_matrix.cols,) + tuple(mat.cols for mat in a_matrices[l + 1:]) + tuple(mat.rows for mat in a_matrices[:l])
        out_shape = (a_matrix.rows,) + tuple(mat.cols for mat in a_matrices[l + 1:]) + tuple(mat.rows for mat in a_matrices[:l])
        out_inames = tuple(sumfact_iname(length, "out_inames_" + str(k)) for k, length in enumerate(out_shape))
        vec_iname = ()
        if a_matrix.vectorized:
            iname = sumfact_iname(4, "vec")
            vec_iname = (prim.Variable(iname),)
            transform(lp.tag_inames, [(iname, "vec")])

        # A trivial reduction is implemented as a product, otherwise we run into
        # a code generation corner case producing way too complicated code. This
        # could be fixed upstream, but the loopy code realizing reductions is not
        # trivial and the priority is kind of low.
        if a_matrix.cols != 1:
            k = sumfact_iname(a_matrix.cols, "red")
            k_expr = prim.Variable(k)
        else:
            k_expr = 0

        # Setup the input of the sum factorization kernel. In the
        # first matrix multiplication this can be taken from
        # * an input temporary (default)
        # * a global data structure (if FastDGGridOperator is in use)
        # * a value from a global data structure, broadcasted to a vector type (vectorized + FastDGGridOperator)
        input_inames = (k_expr,) + tuple(prim.Variable(j) for j in out_inames[1:])
        if l == 0 and direct_input is not None:
            # See comment below
            input_inames = _permute_backward(input_inames, perm)
            inp_shape = _permute_backward(inp_shape, perm)

            globalarg(direct_input, dtype=np.float64, shape=inp_shape, dim_tags=novec_ftags)
            if a_matrix.vectorized:
                input_summand = prim.Call(prim.Variable("Vec4d"),
                                          (prim.Subscript(prim.Variable(direct_input),
                                                          input_inames),))
            else:
                input_summand = prim.Subscript(prim.Variable(direct_input),
                                               input_inames + vec_iname)
        else:
            # If we did permute the order of a matrices above we also
            # permuted the order of out_inames. Unfortunately the
            # order of our input is from 0 to d-1. This means we need
            # to permute _back_ to get the right coefficients.
            if l == 0:
                inp_shape = _permute_backward(inp_shape, perm)
                input_inames = _permute_backward(input_inames, perm)

            # Get a temporary that interprets the base storage of the input
            # as a column-major matrix. In later iteration of the amatrix loop
            # this reinterprets the output of the previous iteration.
            inp = get_buffer_temporary(sf.buffer,
                                       shape=inp_shape + vec_shape,
                                       dim_tags=ftags)

            # The input temporary will only be read from, so we need to silence the loopy warning
            silenced_warning('read_no_write({})'.format(inp))

            input_summand = prim.Subscript(prim.Variable(inp),
                                           input_inames + vec_iname)

        switch_base_storage(sf.buffer)

        # Get a temporary that interprets the base storage of the output.
        #
        # Note: In this step the reordering of the fastest directions
        # is happening. The new direction (out_inames[0]) and the
        # corresponding shape (out_shape[0]) goes to the end (slowest
        # direction) and everything stays column major (ftags->fortran
        # style).
        #
        # If we are in the last step we reverse the permutation.
        output_shape = tuple(out_shape[1:]) + (out_shape[0],)
        if l == len(a_matrices) - 1:
            output_shape = _permute_backward(output_shape, perm)
        out = get_buffer_temporary(sf.buffer,
                                   shape=output_shape + vec_shape,
                                   dim_tags=ftags)

        # Write the matrix-matrix multiplication expression
        matprod = prim.Product((prim.Subscript(prim.Variable(a_matrix.name),
                                               (prim.Variable(out_inames[0]), k_expr) + vec_iname),
                                input_summand))

        # ... which may be a reduction, if k>0
        if a_matrix.cols != 1:
            matprod = lp.Reduction("sum", k, matprod)

        # Here we also move the new direction (out_inames[0]) to the
        # end and reverse permutation
        output_inames = tuple(prim.Variable(i) for i in out_inames[1:]) + (prim.Variable(out_inames[0]),)
        if l == len(a_matrices) - 1:
            output_inames = _permute_backward(output_inames, perm)

        # In case of direct output we directly accumulate the result
        # of the Sumfactorization into some global data structure.
        if l == len(a_matrices) - 1 and direct_output is not None:
            ft = get_global_context_value("form_type")
            if ft == 'residual' or ft == 'jacobian_apply':
                globalarg(direct_output, dtype=np.float64, shape=output_shape, dim_tags=novec_ftags)
                assignee = prim.Subscript(prim.Variable(direct_output), output_inames)
            else:
                assert ft == 'jacobian'
                globalarg(direct_output,
                          dtype=np.float64,
                          shape=output_shape + output_shape,
                          dim_tags=novec_ftags + "," + novec_ftags)
                # TODO: It is at least questionnable, whether using the *order* of the inames in here
                #       for indexing is a good idea. Then again, it is hard to find an alternative.
                _ansatz_inames = tuple(prim.Variable(i) for i in sf.within_inames)
                assignee = prim.Subscript(prim.Variable(direct_output), _ansatz_inames + output_inames)

            # In case of vectorization we need to apply a horizontal add
            if a_matrix.vectorized:
                matprod = prim.Call(prim.Variable("horizontal_add"),
                                    (matprod,))

            # We need to accumulate
            matprod = prim.Sum((assignee, matprod))
        else:
            assignee = prim.Subscript(prim.Variable(out), output_inames + vec_iname)

        # Issue the reduction instruction that implements the multiplication
        # at the same time store the instruction ID for the next instruction to depend on
        insn_dep = frozenset({instruction(assignee=assignee,
                                          expression=matprod,
                                          forced_iname_deps=frozenset([iname for iname in out_inames]).union(frozenset(sf.within_inames)),
                                          forced_iname_deps_is_final=True,
                                          depends_on=insn_dep,
                                          )
                              })

    # Measure times and count operations in c++ code
    if get_option("instrumentation_level") >= 4:
        insn_dep = frozenset({instruction(code="HP_TIMER_STOP({});".format(timer_name),
                                          depends_on=insn_dep,
                                          within_inames=frozenset(sf.within_inames))})
        if sf.stage == 1:
            qp_timer_name = assembler_routine_name() + '_kernel' + '_quadratureloop'
            post_include('HP_DECLARE_TIMER({});'.format(timer_name), filetag='operatorfile')
            dump_accumulate_timer(timer_name)
            insn_dep = instruction(code="HP_TIMER_START({});".format(qp_timer_name),
                                   depends_on=insn_dep)

    if outshape is None:
        assert sf.stage == 3
        outshape = tuple(mat.rows for mat in a_matrices)

    dim_tags = ",".join(['f'] * len(outshape))

    if sf.vectorized:
        outshape = outshape + vec_shape
        # This is a 'bit' hacky: In stage 3 we need to return something with vectag, in stage 1 not.
        if sf.stage == 1:
            dim_tags = dim_tags + ",c"
        else:
            dim_tags = dim_tags + ",vec"

    out = get_buffer_temporary(sf.buffer,
                               shape=outshape,
                               dim_tags=dim_tags,
                               )
    silenced_warning('read_no_write({})'.format(out))

    return next(iter(a_matrices)).output_to_pymbolic(out), insn_dep
